"""Base class to define agent tools."""

from __future__ import annotations

import inspect
from dataclasses import dataclass
from typing import Callable

# Convert Python types to JSON schema types.
TYPE_MAP = {
    "str": "string",
    "int": "number",
    "float": "number",
    "bool": "boolean",
    "list": "array",
    "dict": "object",
}


@dataclass(frozen=True)
class Tool:
    """A tool is a function that can be called by an agent."""

    # The name of the tool.
    name: str

    # A description of how the tool works - this should be detailed
    description: str

    # The function parameters and their types.
    params: tuple[tuple[str, str], ...]

    # The return type of the function.
    return_type: str

    # The function to call.
    fn: Callable

    # Whether the tool needs additional runtime context.
    requires_context: bool = False

    @classmethod
    def from_function(cls, func: Callable) -> Tool:
        """Create a tool from a function.

        Args:
            func: The function to convert to a tool.

        Returns:
            A new tool instance.
        """
        signature = inspect.signature(func)

        # Check for **kwargs parameter
        requires_context = any(
            param.kind == inspect.Parameter.VAR_KEYWORD
            for param in signature.parameters.values()
        )

        # Filter out kwargs from params
        params = tuple(
            (
                name,
                (
                    param.annotation
                    if isinstance(param.annotation, str)
                    else getattr(param.annotation, "__name__", "No annotation")
                ),
            )
            for name, param in signature.parameters.items()
            if param.kind != inspect.Parameter.VAR_KEYWORD
        )

        return_type = (
            signature.return_annotation.__name__
            if hasattr(signature.return_annotation, "__name__")
            else "No annotation"
        )
        description = inspect.getdoc(func) or "No description"
        return cls(
            name=func.__name__,
            description=description,
            params=params,
            return_type=return_type,
            fn=func,
            requires_context=requires_context,
        )


def send_message(message: str) -> None:
    """Send a final message to the user. This should be done after all internal processing is completed.

    Args:
        message: The message to send to the user.
    """
    pass


def list_tools(**kwargs):
    """List all available tools.

    Args:
        kwargs: Additional arguments.
    """
    tools = kwargs.get("messages")
    print(tools)



"""Core agent definitions and functionality for the FlexAI framework.

Defines the Agent class for managing conversations, invoking tools, and
interacting with language models. Provides core functionality for creating
flexible AI agents capable of using various tools and capabilities to assist users.
"""

from __future__ import annotations
import asyncio
import time
from copy import copy
from dataclasses import dataclass, field
from typing import AsyncGenerator, Callable

from flexai.capability import Capability
from flexai.message import (
    UserMessage,
    AIMessage,
    SystemMessage,
    Message,
    MessageContent,
    TextBlock,
    ToolCall,
    ToolResult,
    ToolUseChunk,
)
from flexai.llm import Client, DefaultClient
from flexai.tool import Tool, send_message


@dataclass(frozen=True, kw_only=True)
class AgentConfig:
    """Configuration options for the agent."""


@dataclass(frozen=True, kw_only=True)
class Agent:
    """LLM-powered agent using tools and capabilities to interact with users.

    Manages conversation flow, invokes tools, and leverages a language model
    to generate responses. Supports customization through capabilities and
    a flexible toolset.
    """

    # The system prompt to use for the agent.
    prompt: str | SystemMessage = ""

    # A list of functions that the agent can call and use
    tools: list[Callable] = field(default_factory=list, repr=False)

    # A list of tools that, when called, will terminate the conversation.
    terminating_tools: list[Callable] = field(
        default_factory=lambda: [send_message], repr=False
    )

    # Hooks that can plugin to the main agent loop to modify its behavior.
    capabilities: list[Capability] = field(default_factory=list)

    # The language model to use for the agent.
    llm: Client | None = DefaultClient()

    # Additional configuration options for the agent.
    config: AgentConfig = AgentConfig()

    # The mapping of tool names to tool functions.
    toolbox: dict[str, Tool] = field(default_factory=dict, init=False)

    def __post_init__(self):
        """Perform post-initialization setup."""
        tools = self.tools or []

        # Convert callables to tools and store them in the toolbox.
        tools = {Tool.from_function(tool) for tool in set(tools)}
        tools = {tool.name: tool for tool in tools}

        # Hack around dataclass immutability.
        object.__setattr__(self, "toolbox", tools)

        # Setup the capabilities.
        for capability in self.capabilities:
            capability.setup(self)

    async def modify_messages(
        self, messages: list[Message]
    ) -> AsyncGenerator[MessageContent | Message | list[Message], None]:
        """Hook to modify the messages before sending them to the LLM.

        Args:
            messages: The current conversation messages.

        Yields:
            Intermediate message chunks followed by the modified list of messages.
        """
        # Iterate through the capabilities and modify the messages.
        for capability in self.capabilities:
            print("capability", capability)
            modified_messages = None
            async for output in capability.modify_messages(copy(messages)):
                # This is a partial message chunk.
                if isinstance(output, MessageContent):
                    yield output

                # This is a full message that we want to take note of.
                elif isinstance(output, Message):
                    yield output

                # This is the modified list of messages.
                else:
                    modified_messages = output

            assert modified_messages is not None
            messages = modified_messages

        yield messages

    async def get_system_message(
        self,
    ) -> AsyncGenerator[MessageContent | SystemMessage, None]:
        """Hook to modify the system message before sending it to the LLM.

        Yields:
            Intermediate message chunks followed by the modified system message.
        """
        system = self.prompt
        if isinstance(system, str):
            system = SystemMessage(system)

        # Iterate through the capabilities and modify the system message.
        for capability in self.capabilities:
            async for output in capability.modify_prompt(system):
                # This is a partial message chunk.
                if isinstance(output, MessageContent):
                    yield output

                # This is the modified system message.
                elif isinstance(output, SystemMessage):
                    system = output

        # Cache the system message.
        yield system

    async def modify_response(
        self, messages: list[Message], response: AIMessage
    ) -> AsyncGenerator[MessageContent | AIMessage, None]:
        """Hook to modify the AI response before sending it to the user.

        Args:
            messages: The current conversation messages.
            response: The AI response.

        Yields:
            Intermediate message chunks followed by the modified AI response.
        """
        # Iterate through the capabilities and modify the response.
        for capability in self.capabilities:
            # Our final response
            modified_response = None

            async for output in capability.modify_response(messages, response):
                # This is a partial message chunk.
                if isinstance(output, MessageContent):
                    yield output

                # This is the modified AI response.
                else:
                    # This is a full piece of content (presumably a message) that we want to take note of
                    if modified_response:
                        yield modified_response
                    modified_response = output

            assert modified_response is not None

            response = modified_response

        yield response

    async def invoke_tool(
        self, tool_call: ToolCall, messages: list[Message] = []
    ) -> AsyncGenerator[MessageContent | Message, None]:
        """Execute a tool and yield progress before returning the final result.

        Args:
            tool_call: The tool call to execute.
            messages: The current conversation messages.

        Yields:
            Intermediate message chunks followed by the final tool result.
        """
        tool = self.toolbox[tool_call.name]
        is_error = False

        # Add the context to the tool input.
        kwargs = tool_call.input.copy()
        if tool.requires_context:
            # Add the current messages.
            kwargs["messages"] = messages

            # Add the agent instance.
            kwargs["agent"] = self

        start_time = time.time()
        try:
            if asyncio.iscoroutinefunction(tool.fn):
                result = await tool.fn(**kwargs)
            else:
                result = tool.fn(**kwargs)

            if isinstance(result, AsyncGenerator):
                last_received = None
                async for intermediate in result:
                    if isinstance(intermediate, Message):
                        yield intermediate
                        continue
                    if last_received:
                        yield ToolUseChunk(
                            id=tool_call.id,
                            tool_name=tool_call.name,
                            content=last_received,
                        )
                    last_received = intermediate
                result = last_received

        except Exception as e:
            result = str(e)
            is_error = True

        end_time = time.time()

        yield ToolResult(
            tool_call_id=tool_call.id,
            result=result,
            execution_time=end_time - start_time,
            is_error=is_error,
        )

    async def step(
        self, messages: list[Message], stream: bool = False, allow_tool: bool = True
    ) -> AsyncGenerator[MessageContent | Message | list[Message], None]:
        """Process a single turn in the conversation.

        Generates a response using the language model and determines if any
        tools need to be invoked based on the current conversation state.

        Args:
            messages: Current conversation messages.
            stream: Boolean flag for whether to yield partial message chunks from the API or not.
            allow_tool: Boolean flag for whether to fully interpret tool calls/results in the history.

        Yields:
            Partial chunks of the tool response, if streaming is on.

        Returns:
            The generated responses, including potential tool use messages.

        Raises:
            ValueError: If no LLM client is provided, or if no message is received in stream mode.
        """
        # Ensure an LLM client is provided.
        if self.llm is None:
            raise ValueError("No LLM client provided.")

        # Preprocess the messages and get the system message.
        async for output in self.modify_messages(messages):
            # This is a partial message chunk.
            if isinstance(output, MessageContent):
                yield output

            # This is a full message that we want to take note of
            elif isinstance(output, Message):
                yield output

            # This is the modified list of messages.
            else:
                messages = output
        
        # The runner will want the capability-modified messages
        yield messages

        system = ""
        async for output in self.get_system_message():
            # This is a partial message chunk.
            if isinstance(output, MessageContent):
                yield output

            # This is the modified system message.
            else:
                system = output

        if stream:
            response = None
            async for chunk in self.llm.stream_chat_response(  # type: ignore
                messages, system=system, tools=list(self.toolbox.values()), allow_tool=allow_tool,
            ):
                # This is a partial message chunk.
                if isinstance(chunk, MessageContent):
                    yield chunk

                # This is the final message.
                else:
                    response = chunk
        else:
            # Get the response from the LLM.
            response = await self.llm.get_chat_response(
                messages, system=system, tools=list(self.toolbox.values())
            )

        assert response is not None

        modified_response = None

        # Modify the response.
        async for output in self.modify_response(messages, response):
            # This is a partial message chunk.
            if isinstance(output, MessageContent):
                yield output

            # This is the modified AI response.
            else:
                # This is a full piece of content (presumably a message) that we want to take note of
                if modified_response:
                    yield modified_response
                modified_response = output

        assert modified_response is not None

        response = modified_response

        # Base case: send_message tool call.
        if (
            isinstance(response.content, list)
            and isinstance(response.content[0], ToolCall)
            and any(
                response.content[0].name == tool.__name__
                for tool in self.terminating_tools
            )
        ):
            if response.content[0].name == "send_message":
                response.content = response.content[0].input["message"]
            yield response
            return

        # Return the response.
        yield response

    async def run(
        self, messages: list[Message], stream: bool = False, allow_tool: bool = True
    ) -> AsyncGenerator[MessageContent | Message, None]:
        """Generate an asynchronous stream of agent responses and tool invocations.

        Processes conversation steps and invokes tools until a final response
        (non-tool use message) is generated.

        Args:
            messages: Initial conversation messages.
            stream: Boolean flag for whether to yield partial message chunks from the inner API or not.
            allow_tool: Boolean flag for whether to fully interpret tool calls/results in the history.

        Yields:
            Message: Each message in the conversation, including tool uses and results.

        Returns:
            If we receive a non-tool use message, the function will terminate.
        """
        # Run in a loop.
        while True:
            # Get the response and yield
            response = None
            
            # Obtain the modified messages as well (because some tools might actually want them)
            modified_messages = None
            
            async for output in self.step(messages, stream=stream, allow_tool=allow_tool):
                # This is a partial message chunk.
                if isinstance(output, MessageContent) or isinstance(output, Message):
                    yield output
                
                # This is the modified list of messages.
                if isinstance(output, list) and all(isinstance(message, Message) for message in output):
                    modified_messages = output

                # This *could be* the final message.
                if isinstance(output, Message):
                    response = output

            assert response is not None
            
            assert modified_messages is not None
            
            messages = modified_messages

            # If it's not a tool use, end the stream.
            if isinstance(response.content, str) or isinstance(
                response.content, TextBlock
            ):
                return

            if (
                isinstance(response.content, list)
                and len(response.content) == 1
                and isinstance(response.content[0], TextBlock)
            ):
                return

            results = []
            last_tool_call = None
            for tool_call in response.content:
                if isinstance(tool_call, ToolCall):
                    result = None
                    last_tool_call = tool_call
                    async for intermediate in self.invoke_tool(tool_call, messages):
                        if isinstance(intermediate, MessageContent):
                            yield intermediate
                        if isinstance(intermediate, Message):
                            yield intermediate
                        result = intermediate
                    assert result is not None
                    results.append(result)

            result_message = UserMessage(results)
            yield result_message

            # Append the messages.
            messages.append(response)
            messages.append(result_message)

            if isinstance(last_tool_call, ToolCall) and any(
                last_tool_call.name == tool.__name__ for tool in self.terminating_tools
            ):
                break



"""Module for interacting with different language models."""

from .client import Client as Client

# Import the default client if available.
try:
    from .anthropic import AnthropicClient as DefaultClient
except ImportError:

    def DefaultClient() -> Client | None:
        """Return a default client.

        Returns:
            None since no default client is available.
        """
        return None



"""Capabilities define hooks that can plug into the agent's pipeline."""

from __future__ import annotations

from abc import ABC
from dataclasses import dataclass
from typing import AsyncGenerator, TYPE_CHECKING

from flexai.message import Message, AIMessage, SystemMessage, MessageContent

if TYPE_CHECKING:
    from flexai.agent import Agent


@dataclass
class Capability(ABC):
    """Abstract base class for defining cognitive capabilities of an agent.

    Provides hooks to modify the agent's behavior at different stages of
    the conversation pipeline, allowing for customization of prompts,
    messages, and responses.
    """

    def setup(self, agent: Agent):
        """Perform any setup required by the capability.

        Args:
            agent: The agent that the capability is attached to.
        """
        pass

    async def modify_prompt(
        self, prompt: SystemMessage
    ) -> AsyncGenerator[MessageContent | SystemMessage, None]:
        """Modify the system prompt before it's sent to the LLM.

        Args:
            prompt: The current system prompt.

        Yields:
            Intermediate message chunks followed by the modified system prompt.
        """
        yield prompt

    async def modify_messages(
        self, messages: list[Message]
    ) -> AsyncGenerator[MessageContent | Message | list[Message], None]:
        """Modify the conversation messages before sending them to the LLM.

        This method can be used to add, remove, or alter messages in the
        conversation history before they are processed by the language model.

        Args:
            messages: The current conversation messages.

        Yields:
            Intermediate message chunks followed by the modified list of messages.
        """
        yield messages

    async def modify_response(
        self, messages: list[Message], response: AIMessage
    ) -> AsyncGenerator[MessageContent | AIMessage, None]:
        """Modify the AI-generated response before sending it to the user.

        This method allows for post-processing of the AI's response, which
        can include filtering, reformatting, or augmenting the content.

        Args:
            messages: The current conversation messages.
            response: The AI-generated response.

        Yields:
            Intermediate message chunks followed by the modified AI response.
        """
        yield response



from __future__ import annotations
import json
import os
import time
from dataclasses import dataclass, field
from typing import Any, AsyncGenerator, Type, Sequence

# Try to import the anthropic library.
try:
    from anthropic import AsyncAnthropic  # type: ignore
    from anthropic.types import ToolUseBlock  # type: ignore
    from pydantic import BaseModel  # type: ignore
except ImportError:
    raise ImportError(
        "The anthropic library is required for the AnthropicClient. "
        "Please install it using `pip install anthropic`."
    )
from flexai.llm.client import Client
from flexai.message import (
    AIMessage,
    Message,
    UserMessage,
    SystemMessage,
    ToolCall,
    ToolResult,
    MessageContent,
    TextBlock,
    DataBlock,
    Usage,
)
from flexai.tool import Tool, TYPE_MAP


ANTHROPIC_BETA_HEADERS = [
    "max-tokens-3-5-sonnet-2024-07-15",
]


def get_tool_call(tool_use: ToolUseBlock) -> ToolCall:
    """Get the tool call from a tool use block.

    Args:
        tool_use: The tool use block to get the call from.

    Returns:
        The tool call from the tool use block.
    """
    return ToolCall(
        id=tool_use.id,
        name=tool_use.name,
        input=tool_use.input,
    )


@dataclass(frozen=True)
class AnthropicClient(Client):
    """Client for interacting with the Anthropic language model."""

    # The client to use for interacting with the model.
    client: AsyncAnthropic = AsyncAnthropic()

    # The model to use for generating responses.
    model: str = os.getenv("ANTHROPIC_MODEL", "claude-3-5-sonnet-latest")

    # The maximum number of tokens to generate in a response.
    max_tokens: int = 8192

    # Whether to cache messages or not.
    cache_messages: bool = True

    # Extra headers to include in the request.
    extra_headers: dict = field(
        default_factory=lambda: {"anthropic-beta": ",".join(ANTHROPIC_BETA_HEADERS)}
    )

    async def get_chat_response(
        self,
        messages: list[Message],
        system: str | SystemMessage = "",
        tools: list[Tool] | None = None,
        force_tool: bool = True,
        allow_tool: bool = True,
    ) -> AIMessage:
        # Send the messages to the model and get the response.
        start = time.time()
        params = self._get_params(
            messages=messages,
            system=system,
            tools=tools,
            force_tool=force_tool,
            allow_tool=allow_tool,
            stream=False,
        )
        response = await self.client.messages.create(**params)
        generation_time = time.time() - start

        # Parse out the tool uses from the response.
        tool_uses = [
            get_tool_call(message)
            for message in response.content
            if isinstance(message, ToolUseBlock)
        ]

        # Get the content to return.
        content_to_return = tool_uses or "\n".join(
            [message.text for message in response.content]
        )
        return AIMessage(
            content=content_to_return,
            usage=Usage(
                input_tokens=response.usage.input_tokens,
                output_tokens=response.usage.output_tokens,
                cache_read_tokens=response.usage.cache_read_input_tokens,
                cache_write_tokens=response.usage.cache_creation_input_tokens,
                generation_time=generation_time,
            ),
        )

    async def stream_chat_response(  # type: ignore
        self,
        messages: list[Message],
        system: str | SystemMessage = "",
        tools: list[Tool] | None = None,
        force_tool: bool = True,
        allow_tool: bool = True,
    ) -> AsyncGenerator[MessageContent | AIMessage, None]:
        # Initialize variables to store the tool information.
        tool_call: ToolCall | None = None
        text_block: TextBlock | None = None

        # Track the usage.
        usage = Usage()

        # Iterate over the response stream.
        start = time.time()
        params = self._get_params(
            messages=messages,
            system=system,
            tools=tools,
            force_tool=force_tool,
            allow_tool=allow_tool,
            stream=True,
        )
        response_stream = await self.client.messages.create(**params)
        async for chunk in response_stream:
            # Add the usage.
            # This is the starting chunk.
            if hasattr(chunk, "message"):
                usage.input_tokens += chunk.message.usage.input_tokens
                usage.output_tokens += chunk.message.usage.output_tokens
                usage.cache_read_tokens += chunk.message.usage.cache_read_input_tokens
                usage.cache_write_tokens += (
                    chunk.message.usage.cache_creation_input_tokens
                )
            # This is the continuation chunk.
            if hasattr(chunk, "usage"):
                usage.output_tokens += chunk.usage.output_tokens

            # Content start blocks.
            if chunk.type == "content_block_start":
                content = chunk.content_block
                # This is a text block.
                if content.type == "text":
                    text_block = TextBlock(content.text)
                    yield text_block

                # This is a tool block.
                else:
                    # Yield the initial tool call with no input.
                    tool_call = ToolCall(id=content.id, name=content.name, input="")
                    yield tool_call

            # A continuation of the content block.
            elif chunk.type == "content_block_delta":
                delta = chunk.delta
                # This is a text delta.
                if delta.type == "text_delta":
                    assert text_block is not None
                    text_block = text_block.append(delta.text)
                    yield TextBlock(delta.text)

                # This is a tool delta.
                else:
                    assert tool_call is not None
                    # Add to the input buffer and yield the partial JSON.
                    tool_call = tool_call.append_input(delta.partial_json)
                    yield TextBlock(delta.partial_json)

            # The end of the content block.
            elif chunk.type == "message_stop":
                # Parse the tool from the buffer and convert it to a tool call.
                usage.generation_time = time.time() - start

                # Send the final text message.
                content = None
                if text_block:
                    content = text_block
                    text_block = None
                elif tool_call:
                    # Send the tool call message.
                    content = tool_call.load_input()
                    tool_call = None

                # Send the final message.
                if content:
                    yield AIMessage(
                        content=[content],
                        usage=usage,
                    )

    def _add_cache_control(self, params: dict) -> dict:
        """Add cache control to the params.

        Args:
            params: The params to add cache control to.

        Returns:
            The params with cache control added.
        """
        if not self.cache_messages:
            return params

        cache_control = {
            "cache_control": {
                "type": "ephemeral",
            }
        }

        # Cache the system message.
        params["system"][0].update(**cache_control)

        # Cache tool definitions.
        if "tools" in params and len(params["tools"]) > 0:
            params["tools"][-1].update(**cache_control)

        # Cache the last two user messages.
        user_idxs = [
            idx
            for idx, message in enumerate(params["messages"])
            if message["role"] == "user"
        ]
        for idx in user_idxs[-2:]:
            message = params["messages"][idx]
            if isinstance(message["content"], str):
                message["content"] = [{"type": "text", "text": message["content"]}]
            message["content"][-1].update(**cache_control)

        return params

    def _get_params(
        self,
        messages: list[Message],
        system: str | SystemMessage,
        tools: list[Tool] | None,
        force_tool: bool,
        allow_tool: bool,
        stream: bool,
    ) -> dict:
        """Get the common params to send to the model.

        Args:
            messages: The messages to send to the model.
            system: The system message to send to the model.
            tools: The tools to send to the model.
            force_tool: Whether to force the model to use the tools.
            allow_tool: Whether to allow tool calls in the content.
            stream: Whether to stream the response.

        Returns:
            The common params to send to the model.
        """
        # Convert the system prompt to a list of message content.
        if isinstance(system, str):
            system = SystemMessage([TextBlock(system)])

        kwargs = {
            "max_tokens": self.max_tokens,
            "messages": self.format_content(messages, allow_tool),
            "model": self.model,
            "system": self.format_content(system.content, allow_tool),
            "extra_headers": self.extra_headers,
        }

        # If tools are provided, force the model to use them (for now).
        if tools:
            kwargs["tools"] = [self.format_tool(tool) for tool in tools]
            if force_tool:
                kwargs["tool_choice"] = {"type": "any"}
        else:
            kwargs["tools"] = []

        if stream:
            kwargs["stream"] = True

        # Add cache control to the params.
        params = self._add_cache_control(kwargs)
        return params

    async def get_structured_response(
        self,
        messages: list[Message],
        model: Type[BaseModel],
        system: str | SystemMessage = "",
        tools: list[Tool] | None = None,
    ) -> BaseModel:
        """Get the structured response from the chat model.

        Args:
            messages: The messages to send to the model.
            model: The model to use for the response.
            system: Optional system message to set the behavior of the AI.
            tools: Tools to use in the response.

        Returns:
            The structured response from the model.

        Raises:
            ValueError: If the response is not a string.
        """
        schema = model.schema()  # type: ignore
        system = f"""{system}
Return your answer according to the 'properties' of the following schema:
{schema}
Return only the JSON object with the properties filled in.
Do not include anything in your response other than the JSON object.
Do not begin your response with ```json or end it with ```.
"""
        response = await self.get_chat_response(
            messages, system=system, tools=tools, force_tool=False
        )
        response.content = str(response.content)
        content = response.content
        try:
            if not isinstance(content, str):
                raise ValueError("The response is not a string.")
            return model.parse_raw(content)  # type: ignore
        except Exception as e:
            # Try again, printing the exception.
            messages = messages + [
                response,
                UserMessage(
                    f"There was an error while parsing. Make sure to only include the JSON. Error: {e}"
                ),
            ]
            return await self.get_structured_response(
                messages, model=model, system=system, tools=tools
            )

    @staticmethod
    def format_tool(tool: Tool) -> dict:
        """Convert the tool to a description.

        Args:
            tool: The tool to format.

        Returns:
            A dictionary describing the tool.
        """
        input_schema = {
            "type": "object",
            "properties": {},
        }
        for param_name, param_type in tool.params:
            param_type = TYPE_MAP.get(str(param_type), param_type)
            input_schema["properties"][param_name] = {
                "type": param_type,
            }

        description = {
            "name": tool.name,
            "description": tool.description,
            "input_schema": input_schema,
        }
        return description

    @classmethod
    def format_content(
        cls,
        value: str
        | Message
        | MessageContent
        | Sequence[MessageContent]
        | Sequence[Message],
        allow_tool: bool = True,
    ) -> Any:
        """Format the message content for the Anthropic model.

        Args:
            value: The value to format.
            allow_tool: Whether to allow tool calls in the content.

        Returns:
            The formatted message content.

        Raises:
            ValueError: If the message content type is unknown.
        """
        # Base types.
        if isinstance(value, str):
            return value
        if isinstance(value, list):
            return [cls.format_content(v, allow_tool) for v in value]

        # Anthropic message format.
        if isinstance(value, Message):
            if isinstance(value.content, ToolCall):
                raise ValueError("Tool calls should not be included in Messages.")
            return {
                "role": value.role,
                "content": cls.format_content(value.content, allow_tool),
            }

        # Message content types.
        if isinstance(value, TextBlock):
            return {
                "type": "text",
                "text": value.text,
            }
        if isinstance(value, DataBlock):
            return {
                "type": "text",
                "text": json.dumps(value.data),
            }
        if isinstance(value, ToolCall):
            if allow_tool:
                return {
                    "type": "tool_use",
                    "id": value.id,
                    "name": value.name,
                    "input": value.input,
                }
            else:
                return {
                    "type": "text",
                    "text": f"Tool Use\nid:{value.id}\nname:{value.name}\ninput:{json.dumps(value.input)}",
                }
        if isinstance(value, ToolResult):
            if allow_tool:
                return {
                    "type": "tool_result",
                    "tool_use_id": value.tool_call_id,
                    "content": str(value.result),
                    "is_error": value.is_error,
                }
            else:
                return {
                    "type": "text",
                    "text": f"Tool Result\ntool_call_id: {value.tool_call_id}\nresult:{value.result}\nexecution:{value.execution_time}\nis_error:{value.is_error}",
                }

        raise ValueError(f"Unknown message content type: {value}")

    @classmethod
    def load_content(
        cls, content: str | list[dict[str, Any]]
    ) -> str | list[MessageContent]:
        """Load the message content from the Anthropic model to dataclasses.

        Args:
            content: The content to load.

        Returns:
            The loaded message content
        """
        # If it's a string, return it.
        if isinstance(content, str):
            return content

        # If it's a list of dictionaries, parse them.
        assert isinstance(content, list)
        parsed_content: list[MessageContent] = []

        for entry in content:
            match entry.pop("type"):
                case "text":
                    parsed_content.append(TextBlock(**entry))
                case "data":
                    parsed_content.append(DataBlock(**entry))
                case "tool_use":
                    parsed_content.append(ToolCall(**entry))
                case "tool_result":
                    parsed_content.append(
                        ToolResult(
                            tool_call_id=entry.pop("tool_use_id"),
                            result=entry.pop("content"),
                            **entry,
                        )
                    )

        return parsed_content



from dataclasses import dataclass
from typing import AsyncGenerator

from flexai.capability import Capability
from flexai.message import Message, MessageContent


@dataclass
class TruncateMessages(Capability):
    """Truncate the input messages to the LLM to a maximum number."""

    # The maximum number of messages to keep.
    max_messages: int

    async def modify_messages(
        self, messages: list[Message]
    ) -> AsyncGenerator[MessageContent | list[Message], None]:
        yield messages[-self.max_messages :]



"""A library to build AI agents with flexible capabilities and tool use."""

from .agent import Agent as Agent
from .capability import Capability as Capability
from .message import (
    AIMessage as AIMessage,
    Message as Message,
    UserMessage as UserMessage,
)



from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import AsyncGenerator

from flexai.message import AIMessage, Message, MessageContent, SystemMessage
from flexai.tool import Tool


@dataclass(frozen=True)
class Client(ABC):
    """Abstract base class for language model clients.

    Defines the interface for interacting with various language models.
    Subclasses should implement the necessary methods for specific LLM providers.
    """

    @abstractmethod
    async def get_chat_response(
        self,
        messages: list[Message],
        system: str | SystemMessage = "",
        tools: list[Tool] | None = None,
    ) -> AIMessage:
        """Retrieve a response from the chat model.

        Args:
            messages: Conversation history to send to the model.
            system: Optional system message to set the behavior of the AI.
            tools: Optional list of tools available for the model to use.

        Returns:
            A list of AI-generated messages in response to the input.
        """

    @abstractmethod
    async def stream_chat_response(
        self,
        messages: list[Message],
        system: str | SystemMessage = "",
        tools: list[Tool] | None = None,
        allow_tool: bool = True,
    ) -> AsyncGenerator[MessageContent | AIMessage, None]:  # type: ignore
        """Stream the response from the chat model in real-time.

        Args:
            messages: Conversation history to send to the model.
            system: Optional system message to set the behavior of the AI.
            tools: Optional list of tools available for the model to use.
            allow_tool: Boolean flag for whether to fully interpret tool calls/results in the history.

        Yields:
            AI-generated messages as they are generated by the model.
        """



from dataclasses import dataclass
from typing import AsyncGenerator

from flexai.capability import Capability
from flexai.message import Message, AIMessage, MessageContent, ToolCall
import json


@dataclass
class LimitToolUse(Capability):
    """Force an agent to stop if there are too many tool use."""

    # The maximum number of tool uses allowed.
    max_tool_uses: int

    def is_tool_call(self, entry):
        return isinstance(entry, ToolCall) or (
            isinstance(entry, dict) and entry["type"] == "tool_use"
        )

    async def modify_response(
        self, messages: list[Message], response: AIMessage
    ) -> AsyncGenerator[MessageContent | AIMessage, None]:
        # There's nothing to modify
        if not isinstance(response.content, list):
            yield response
            return

        # Initial: Number of tool uses that this current response requests.
        total_tool_uses = len(
            [entry for entry in response.content if self.is_tool_call(entry)]
        )

        for message in messages:
            try:
                if not isinstance(message.content, str):
                    raise ValueError("Content is not a string")
                content = json.loads(message.content)
            except Exception:
                content = message.content

            if not isinstance(content, list):
                continue

            # For each past message, add the number of tool uses it suggested to our total count.
            for entry in content:
                if self.is_tool_call(entry):
                    total_tool_uses += 1

        # If the total tool use count, including this response, is too much, we can't use one more
        if total_tool_uses > self.max_tool_uses:
            send_message_call = ToolCall(
                id="",
                name="send_message",
                input={"message": f"Exceeded tool usage limit: {self.max_tool_uses}"},
            )
            response.content = [send_message_call]

        yield response



"""Message types for agent-LLM communication in the FlexAI framework."""

from __future__ import annotations
from dataclasses import asdict, dataclass, field, replace
from typing import Any, Sequence
import json


@dataclass
class Message:
    """Base class for all message types in the conversation flow."""

    # The role of the message (user or assistant).
    role: str

    # The content of the message.
    content: str | Sequence[MessageContent]

    @classmethod
    def load(cls, data: dict[str, Any]) -> Message:
        return cls(
            role=data["role"],
            content=data["content"]
            if isinstance(data["content"], str)
            else [MessageContent.load(content) for content in data["content"]],
        )

    def dump(self):
        return {
            "role": self.role,
            "content": self.content
            if isinstance(self.content, str)
            else [content.dump() for content in self.content],
        }


@dataclass
class SystemMessage(Message):
    """A top level system message."""

    role: str = field(init=False, default="system")


@dataclass
class UserMessage(Message):
    """A message sent by a user."""

    role: str = field(init=False, default="user")


@dataclass
class AIMessage(Message):
    """A message generated by the AI."""

    # The resource usage for the message.
    role: str = field(init=False, default="assistant")

    # The resource usage for the message.
    usage: Usage = field(default_factory=lambda: Usage())


@dataclass(frozen=True)
class MessageContent:
    def dump(self) -> dict[str, Any]:
        """Return the dataclass as a dictionary.

        Returns:
            The dataclass as a dictionary.
        """
        return {"type": self.__class__.__name__, **asdict(self)}

    @classmethod
    def load(cls, data: dict[str, Any]) -> MessageContent:
        """Load the dataclass from a dictionary.

        Args:
            data: The dictionary to load.

        Returns:
            The dataclass instance.

        Raises:
            ValueError: If the message content type is unknown.
        """
        # Check all subclasses of MessageContent.
        for subclass in cls.__subclasses__():
            if subclass.__name__ == data["type"]:
                data = {k: v for k, v in data.items() if k != "type"}
                return subclass(**data)

        raise ValueError(f"Unknown message content type: {data['type']}")


@dataclass(frozen=True)
class TextBlock(MessageContent):
    """A block of text content."""

    text: str

    def append(self, text: str) -> TextBlock:
        """Append text to the block.

        Args:
            text: The text to append.

        Returns:
            A new TextBlock instance with the appended text.
        """
        return replace(self, text=self.text + text)


@dataclass(frozen=True)
class DataBlock(MessageContent):
    """A block of structured data."""

    data: dict[str, Any]


@dataclass(kw_only=True, frozen=True)
class ToolUseChunk(MessageContent):
    """A chunk reported by the tool during execution."""

    # A unique identifier for the tool call this came from.
    id: str

    # The name of the tool this is from.
    tool_name: str

    # Associated data.
    content: Any


@dataclass(kw_only=True, frozen=True)
class ToolCall(MessageContent):
    """A tool call message sent by the agent."""

    # A unique identifier for the tool call.
    id: str

    # The name of the tool to call.
    name: str

    # The input parameters for the tool.
    input: Any

    def append_input(self, chunk: str) -> ToolCall:
        """Append a chunk to the input data.

        Args:
            chunk: The chunk of text to append from the stream.

        Returns:
            A new ToolCall instance with the appended input.
        """
        assert isinstance(self.input, str)
        return replace(self, input=self.input + chunk)

    def load_input(self) -> ToolCall:
        """Load the input string as JSON.

        Returns:
            A new ToolCall instance with the input loaded as JSON.
        """
        assert isinstance(self.input, str)
        try:
            data = json.loads(self.input)
        except json.JSONDecodeError:
            data = {}
        return replace(self, input=data)


@dataclass(kw_only=True, frozen=True)
class ToolResult(MessageContent):
    """A tool result message created after invoking a tool."""

    # The associated tool call identifier.
    tool_call_id: str

    # The result of the tool invocation.
    result: Any

    # The execution time of the tool invocation.
    execution_time: float = 0.0

    # Whether an error occurred during invocation.
    is_error: bool = False


@dataclass(kw_only=True)
class Usage:
    """Track resource usage for a message."""

    # The number of input tokens used to generate the message.
    input_tokens: int = 0

    # The number of output tokens used to generate the message.
    output_tokens: int = 0

    # The number of tokens read from the cache.
    cache_read_tokens: int = 0

    # The number of tokens written to the cache.
    cache_write_tokens: int = 0

    # The time taken to generate the message.
    generation_time: float = 0.0



from __future__ import annotations
import json
import itertools
import os
import time
from dataclasses import dataclass, field, InitVar
from typing import Any, AsyncGenerator, Sequence, Type

# Try to import the openai library.
try:
    from openai import AsyncOpenAI  # type: ignore
    from pydantic import BaseModel  # type: ignore
except ImportError:
    raise ImportError(
        "The openai library is required for the OpenAIClient. "
        "Please install it using `pip install openai`."
    )
from flexai.llm.client import Client
from flexai.message import (
    AIMessage,
    Message,
    ToolCall,
    ToolResult,
    MessageContent,
    SystemMessage,
    DataBlock,
    TextBlock,
    Usage,
)
from flexai.tool import Tool, TYPE_MAP


def get_tool_call(tool_use) -> ToolCall:
    """Get the tool call from a tool use block.

    Args:
        tool_use: The tool use block to get the call from.

    Returns:
        The tool call from the tool use block.
    """
    return ToolCall(
        id=tool_use.id,
        name=tool_use.function.name,
        input=json.loads(tool_use.function.arguments),
    )


@dataclass(frozen=True)
class OpenAIClient(Client):
    """Client for interacting with the Anthropic language model."""

    # The API key to use for interacting with the model.
    api_key: InitVar[str] = field(default=os.environ.get("OPENAI_API_KEY", ""))

    # The base URL to use for interacting with the model.
    base_url: InitVar[str] = field(default="https://api.openai.com/v1")

    # The client to use for interacting with the model.
    client: AsyncOpenAI = AsyncOpenAI()

    # The model to use for generating responses.
    model: str = os.getenv("OPENAI_MODEL", "gpt-4o-mini")

    def __post_init__(self, api_key, base_url):
        object.__setattr__(
            self, "client", AsyncOpenAI(api_key=api_key, base_url=base_url)
        )

    async def get_chat_response(
        self,
        messages: list[Message],
        system: str | SystemMessage = "",
        tools: list[Tool] | None = None,
        force_tool: bool = True,
        allow_tool: bool = True,
    ) -> AIMessage:
        # Send the messages to the model and get the response.
        params = self._get_params(
            messages=messages,
            system=system,
            tools=tools,
            force_tool=force_tool,
            allow_tool=allow_tool,
            stream=False,
        )
        start = time.time()
        response = await self.client.chat.completions.create(**params)
        generation_time = time.time() - start

        # Parse out the tool uses from the response.
        message = response.choices[0].message
        tool_uses = [get_tool_call(message) for message in message.tool_calls]

        # Get the content to return.
        content_to_return = tool_uses or message.content
        return AIMessage(
            content=content_to_return,
            usage=Usage(
                input_tokens=response.usage.prompt_tokens,
                output_tokens=response.usage.completion_tokens,
                generation_time=generation_time,
            ),
        )

    async def stream_chat_response(  # type: ignore
        self,
        messages: list[Message],
        system: str | SystemMessage = "",
        tools: list[Tool] | None = None,
        force_tool: bool = True,
        allow_tool: bool = True,
    ) -> AsyncGenerator[TextBlock | AIMessage, None]:
        stream = await self.client.chat.completions.create(
            **self._get_params(
                messages=messages,
                system=system,
                tools=tools,
                force_tool=force_tool,
                allow_tool=allow_tool,
                stream=True,
            )
        )
        # Currently, this guy only interprets string messages. Which sufficiently contains our use case.
        message_content_buffer = ""
        async for chunk in stream:
            content = chunk.choices[0].delta.content
            if content:
                yield TextBlock(text=content)
                message_content_buffer += content
        yield AIMessage(content=[TextBlock(text=message_content_buffer)])

    def _get_params(
        self,
        messages: list[Message],
        system: str | SystemMessage,
        tools: list[Tool] | None,
        force_tool: bool = False,
        allow_tool: bool = False,
        stream: bool = False,
    ) -> dict:
        """Get the common params to send to the model.

        Args:
            messages: The messages to send to the model.
            system: The system message to send to the model.
            tools: The tools to send to the model.
            force_tool: unused
            allow_tool: unused
            stream: Whether we are streaming.

        Returns:
            The common params to send to the model.
        """
        if isinstance(system, str):
            system = SystemMessage(system)
        messages = self.format_content([system, *messages])
        kwargs = {
            "model": self.model,
            "messages": messages,
        }

        # If tools are provided, force the model to use them (for now).
        if tools:
            kwargs["tools"] = [self.format_tool(tool) for tool in tools]
            kwargs["tool_choice"] = "required"

        if stream:
            kwargs["stream"] = True

        return kwargs

    async def get_structured_response(
        self,
        messages: list[Message],
        model: Type[BaseModel],
        system: str | SystemMessage = "",
        tools: list[Tool] | None = None,
    ) -> BaseModel:
        """Get the structured response from the chat model.

        Args:
            messages: The messages to send to the model.
            model: The model to use for the response.
            system: Optional system message to set the behavior of the AI.
            tools: unused.

        Returns:
            The structured response from the model.

        Raises:
            ValueError: If the response cannot be parsed.
        """
        # Send the messages to the model and get the response.
        response = await self.client.beta.chat.completions.parse(  # type: ignore
            **self._get_params(messages, system, []),
            response_format=model,
        )
        result = response.choices[0].message.parsed
        if result is None:
            raise ValueError("Failed to parse the response.")
        return result

    @staticmethod
    def format_tool(tool: Tool) -> dict:
        """Convert the tool to a description.

        Args:
            tool: The tool to format.

        Returns:
            A dictionary describing the tool.
        """
        input_schema = {
            "type": "object",
            "properties": {},
        }
        for param_name, param_type in tool.params:
            param_type = TYPE_MAP.get(str(param_type), param_type)
            input_schema["properties"][param_name] = {
                "type": param_type,
            }

        description = {
            "type": "function",
            "function": {
                "name": tool.name,
                "description": tool.description,
                "parameters": input_schema,
            },
        }
        return description

    @classmethod
    def format_content(
        cls,
        value: str
        | Message
        | MessageContent
        | Sequence[MessageContent]
        | Sequence[Message],
        allow_tool: bool = True,
    ) -> Any:
        """Format the message content for the Anthropic model.

        Args:
            value: The value to format.
            allow_tool: unused.

        Returns:
            The formatted message content.

        Raises:
            ValueError: If the message content type is unknown.
        """
        # Base types.
        if isinstance(value, str):
            return value
        if isinstance(value, list):
            output = [cls.format_content(v) for v in value]
            # Flatten output messages
            return list(
                itertools.chain.from_iterable(
                    item if isinstance(item, list) else [item] for item in output
                )
            )

        # Anthropic message format.
        if isinstance(value, Message):
            if isinstance(value.content, list):
                if value.role == "user":
                    # Tool calls use a "tool" role instead of "user".
                    messages = []
                    # OpenAI wants ToolResults, which currently are simply items in value.content, to be their own messages
                    # Following is logic to extract those into their own messages, and then format the remainder of value.content as a single entity
                    content = []
                    for item in value.content:
                        if isinstance(item, ToolResult):
                            messages.append(cls.format_content(item))
                        else:
                            content.append(item)
                    messages.append(
                        {
                            "role": value.role,
                            "content": cls.format_content(content),
                        }
                    )
                    return messages

                if value.role == "system":
                    return {
                        "role": value.role,
                        "content": json.dumps(cls.format_content(value.content)),
                    }

                # Assistants use lists for tool calls.
                return {
                    "role": value.role,
                    "tool_calls": [cls.format_content(v) for v in value.content],
                }
            else:
                return {
                    "role": value.role,
                    "content": cls.format_content(value.content),
                }

        # Message content types.
        if isinstance(value, TextBlock):
            return {"type": "text", "text": value.text}
        if isinstance(value, DataBlock):
            return {"type": "text", "text": json.dumps(value.data)}
        if isinstance(value, ToolCall):
            return {
                "type": "function",
                "id": value.id,
                "function": {
                    "name": value.name,
                    "arguments": json.dumps(value.input),
                },
            }
        if isinstance(value, ToolResult):
            return {
                "role": "tool",
                "tool_call_id": value.tool_call_id,
                "content": json.dumps(value.result),
            }
        raise ValueError(f"Unknown message content type: {type(value)}")

    @classmethod
    def load_content(
        cls, content: str | list[dict[str, Any]]
    ) -> str | list[MessageContent]:
        """Load the message content from the Anthropic model to dataclasses.

        Args:
            content: The content to load.

        Returns:
            The loaded message content
        """
        # If it's a string, return it.
        if isinstance(content, str):
            return content

        # If it's a list of dictionaries, parse them.
        assert isinstance(content, list)
        parsed_content: list[MessageContent] = []

        for entry in content:
            match entry.pop("type"):
                case "text":
                    parsed_content.append(TextBlock(**entry))
                case "function":
                    parsed_content.append(
                        ToolCall(
                            id=entry["id"],
                            name=entry["function"]["name"],
                            input=json.loads(entry["function"]["arguments"]),
                        )
                    )
                case "tool":
                    parsed_content.append(
                        ToolResult(
                            tool_call_id=entry.pop("tool_call_id"),
                            result=json.loads(entry.pop("content")),
                            **entry,
                        )
                    )

        return parsed_content
